#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np
import pygame

from ..utils import bounds
from .utils import *
from .video import Video

pygame.init()


class Timeline:
    """
    Frame browser area.
    """

    def __init__(self, video: Video):
        self.video = video

        self.frame = 0

    def draw(self, surface: pygame.Surface, events, rect):
        """
        Draw timeline on surface.

        :param rect: ``(x, y, w, h)`` coordinates to draw.
        """
        x, y, w, h = rect
        total_frames = self.video.num_frames

        # Handle events
        if pygame.mouse.get_pressed()[0]:
            mx, my = pygame.mouse.get_pos()
            if x <= mx < x+w and y <= my < y+h:
                fac = (mx-x) / w
                self.frame = int(fac * self.video.num_frames)

        pygame.draw.rect(surface, DARK_GRAY, rect)

        # Draw cached
        x2 = Timeline.fac_to_pix(x, w, self.video.extracted/total_frames)
        pygame.draw.rect(surface, GRAY, (x, y, x2-x, 5))

        # Draw marker
        marker = Timeline.fac_to_pix(x, w, self.frame/total_frames)
        pygame.draw.line(surface, BLUE, (marker, y), (marker, y+h))

    def next_frame(self):
        self.frame = bounds(self.frame+1, 0, self.video.num_frames)

    def prev_frame(self):
        self.frame = bounds(self.frame-1, 0, self.video.num_frames)

    @staticmethod
    def fac_to_pix(x, w, fac):
        return np.interp(fac, (0, 1), (x, x+w))



#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
import random

import numpy as np

from ..cpp import Types
from .effect import Effect


class Glare(Effect):
    """
    Light glare when notes play.
    """

    def __init__(self, settings, cache, libs, notes) -> None:
        super().__init__(settings, cache, libs)

        for note in notes:
            streaks = []
            for _ in range(settings.glare.streaks):
                angle = random.randint(0, 255)
                streaks.append(angle)
            note.attrs["glare.streak_angles"] = streaks

    def render(self, settings, img: np.ndarray, frame: int, notes):
        """
        Render the glare.

        :param notes: MIDI notes from parse_midi.
        """
        keys = np.array([n.note for n in notes], dtype=Types.int)
        starts = np.array([n.start for n in notes], dtype=Types.double)
        ends = np.array([n.end for n in notes], dtype=Types.double)

        angles = []
        for note in notes:
            angles.extend(note.attrs["glare.streak_angles"])
        angles = np.array(angles, dtype=np.uint8)

        cache_path = os.path.join(self.cache, "glare", "streaks.bin")

        settings = self.settings
        settings_args = [settings.piano.black_width_fac,
            settings.glare.radius, settings.glare.intensity,
            settings.glare.jitter, settings.glare.streaks]

        self.libs["glare"].render_glare(
            img, img.shape[1], img.shape[0],
            frame,
            len(notes), keys, starts, ends, angles,
            *settings_args,
        )



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np
import pygame

pygame.init()

TMP = "/tmp"

BLACK = np.array((0, 0, 0))
DARK_GRAY = np.array((50, 50, 50))
GRAY = np.array((120, 120, 120))
LIGHT_GRAY = np.array((170, 170, 170))
WHITE = np.array((255, 255, 255))
BLUE = np.array((130, 130, 255))

FONT_MED = pygame.font.SysFont("ubuntu", 18)



#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import cv2
import pygame

from .timeline import Timeline
from .utils import *
from .video import Video

pygame.init()


class Preview:
    """
    Preview current frame.
    """

    def __init__(self, video: Video, timeline: Timeline):
        self.video = video
        self.timeline = timeline

    def draw(self, surface: pygame.Surface, events, rect):
        """
        Draw info on surface.

        :param rect: ``(x, y, w, h)`` coordinates to draw.
        """
        x, y, w, h = rect

        if self.timeline.frame > self.video.extracted:
            return

        img = self.video.get(self.timeline.frame)
        img = cv2.resize(img, (w, h)).swapaxes(0, 1)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = pygame.surfarray.make_surface(img)
        surface.blit(img, (x, y))



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import ctypes
import os
from typing import Mapping

from ..cpp import build_lib, Types


def load_libs(cache: str) -> Mapping[str, ctypes.CDLL]:
    """
    Load C libraries.
    """
    cache = os.path.join(cache, "c_libs")
    os.makedirs(cache, exist_ok=True)

    blocks = build_lib(
        ["effects/blocks.cpp"],
        cache,
        "blocks",
    )
    blocks.render_blocks.argtypes = [
        Types.img, Types.int, Types.int,
        Types.int,
        Types.int, Types.arr_int, Types.arr_double, Types.arr_double,
        Types.int, Types.double, Types.double, Types.double, Types.arr_uchar,
    ]

    glare = build_lib(
        ["effects/glare.cpp"],
        cache,
        "glare",
    )
    glare.render_glare.argtypes = [
        Types.img, Types.int, Types.int,
        Types.int,
        Types.int, Types.arr_int, Types.arr_double, Types.arr_double,
            Types.arr_uchar,
        Types.double, Types.double, Types.double, Types.double, Types.int,
    ]

    return {
        "blocks": blocks,
        "glare": glare,
    }



#
#  PianoRay
#  Piano performance visualizer.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import ctypes
from typing import Mapping

import numpy as np

from ..settings import Settings


class Effect:
    """
    Base class for all effects.
    """

    settings: Settings
    cache: str
    libs: Mapping[str, ctypes.CDLL]

    def __init__(self, settings: Settings, cache: str, libs) -> None:
        self.settings = settings
        self.cache = cache
        self.libs = libs

    def render(self, settings: Settings, img: np.ndarray,
            frame: int, *args, **kwargs) -> None:
        """
        Override this in the subclass.
        """
        raise NotImplementedError("Override Effect.render")



#
#  PianoRay
#  Video rendering pipeline with piano visualization.
#  Copyright  PianoRay Authors  2022
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import cv2
import numpy as np

from .effect import Effect


class VideoRead:
    """
    Read frames of a video.
    The class internally accommodates for FPS.
    """
    # Video fps is in_fps, client fps is out_fps
    in_fps: int
    out_fps: int

    def __init__(self, path: str, fps: int, offset: int = 0):
        """
        Initialize.

        :param path: Path of video.
        :param fps: The FPS the client is rendering at.
            i.e. settings.video.fps
        :param offset: Timestamp, in seconds, of the frame that will
            be considered frame 0.
        """
        self._video = cv2.VideoCapture(path)

        self.in_fps = self._video.get(cv2.CAP_PROP_FPS)
        self.out_fps = fps

        # _frame stores frame number with first note = 0
        # _real_frame stores frame number with first frame of video=0
        self._frame = int(-1 * offset * self.in_fps)
        self._real_frame = 0
        self._last = None

    def _get_frame(self, frame: int) -> int:
        """
        Return frame of input video corresponding to
        client's frame.
        """
        f = frame * self.in_fps / self.out_fps
        return round(f)

    def _read_next(self, inc: bool = True):
        """
        Read next frame, store in self._last, and
        increment self._frame.

        :param inc: Whether to increment.
        """
        ret, img = self._video.read()
        self._real_frame += 1
        if inc:
            self._frame += 1
        if ret:
            self._last = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    def read(self, frame: int) -> np.ndarray:
        """
        Read frame. Pass the frame the client needs.
        Currently can only read monotonically.
        """
        f = self._get_frame(frame)
        if f < self._frame:
            raise ValueError("VideoRead can only read monotonically.")

        while self._frame < f:
            self._read_next()

        return self._last


class Keyboard(Effect):
    """
    Piano keyboard rendering.
    """

    def __init__(self, settings, cache, libs) -> None:
        assert settings.keyboard.file is not None

        super().__init__(settings, cache, libs)
        self.video = VideoRead(settings.keyboard.file,
            settings.video.fps, settings.keyboard.start)

        self.compute_crop(settings)

    @staticmethod
    def extend_vec(v1, v2, length):
        """
        Let u be the unit vector in the direction from v1 to v2.
        Let v be u * length.
        Returns v + v2
        """
        diff = v2 - v1
        u = diff / np.linalg.norm(diff)
        v = u * length
        return v + v2

    def compute_crop(self, settings):
        crop = np.array(settings.keyboard.crop)
        src_width = np.linalg.norm(crop[1]-crop[0])
        dst_width = settings.video.resolution[0]
        scale = src_width / dst_width

        below_len = settings.keyboard.below_length * scale
        src_points = np.array(
            (
                crop[0],
                crop[1],
                self.extend_vec(crop[1], crop[2], below_len),
                self.extend_vec(crop[0], crop[3], below_len),
            ),
            dtype=np.float32)

        dst_width = settings.video.resolution[0]
        dst_height = np.linalg.norm(src_points[3]-src_points[0]) / scale
        dst_kbd_height = np.linalg.norm(crop[3]-crop[0]) / scale
        dst_width, dst_height, dst_kbd_height = \
            map(int, (dst_width, dst_height, dst_kbd_height))
        dst_points = np.array(
            ((0,0), (dst_width,0), (dst_width,dst_height), (0,dst_height)),
            dtype=np.float32)

        mask = np.empty((dst_height, dst_width, 3), dtype=np.float32)
        for y in range(dst_height):
            fac = 1 if y < dst_kbd_height else \
                np.interp(y, (dst_kbd_height, dst_height), (1, 0))
            mask[y, :] = fac

        self.persp = cv2.getPerspectiveTransform(src_points, dst_points)
        self.dst_shape = (dst_width, dst_height)
        self.mask = mask

    def render(self, settings, img: np.ndarray, frame: int):
        """
        Render the keyboard.
        """
        dst = self.dst_shape

        kbd = self.video.read(frame)
        kbd = cv2.warpPerspective(kbd, self.persp, dst).astype(np.float64)
        kbd *= self.mask

        kbd *= settings.keyboard.dim_mult
        kbd += settings.keyboard.dim_add
        kbd[kbd<0] = 0
        kbd[kbd>255] = 255
        kbd = kbd.astype(np.uint8)

        half = int(settings.video.resolution[1] / 2)
        img[half:half+dst[1], 0:dst[0], ...] = kbd
